#include "os.h"

#ifndef CONFIGDIR
#define CONFIGDIR "."
#endif

#ifdef HAVE_SYSLOG_H
#define LOG_FACILITY LOG_LOCAL7
#else
#define LOG_FACILITY 0
#endif

#define INDEXER_USER_AGENT	"UdmSearch"
#define MAXDOCSIZE		1024*1024  /* 1Mb */
#define MAXWORDSIZE		32
#define MAXURLSIZE		127
#define MAXLANGPERDOC		16

/* Max URLs in cache */
#define MAXHSIZE		1023

/* Strings size */
#define STRSIZ			1024*5

/* Some usefull utilities */
extern char * gettoken(char *s, const char *delim, char **last);
extern char * trim_chars(char *p, char * delim);
extern char * rtrim_chars(char* p, char *delim);
#define free_it(x)		{if(x){free(x);x=NULL;}}
#define STRNCMP(x,y)		(strncmp(x,y,strlen(y)))
#define STRNCASECMP(x,y)	(strncasecmp(x,y,strlen(y)))
#define SKIP(s,set)		while((*s)&&(strchr(set,*s)))s++;
#define SKIPN(s,set)		while((*s)&&(!strchr(set,*s)))s++;
#define STREND(s)		(s+strlen(s))
#define STRNCPY(x,y)		{strncpy(x,y,sizeof(x));x[sizeof(x)-1]=0;}



/* Common constants */

#define ORD_RATE		1
#define ORD_DATE		2
#define MOD_ALL			1
#define MOD_ANY			2

#define IND_UNKNOWN		0
#define IND_OK			1
#define IND_ERROR		2
#define IND_NO_TARGET		3
#define IND_TERMINATED		4

#define FLAG_REINDEX		1
#define FLAG_EXP_FIRST		2
#define FLAG_ADD_SERV		4
#define FLAG_MARK		8
#define FLAG_SPELL		16



/* UdmSearch types */


/* StopList unit */
typedef struct stopword_struct {
	char *word;
	char lang[3];
} STOP_WORD;


/* Word list unit */
typedef struct word_struct {
	int count;
	int code;
	char *word;
} UDM_WORD;


/* Language quesser structure for one language */
typedef struct lang_struct {
	char lang[3];
	int  count;
} UDM_LANG;


/* All links are stored in the cache of this structure */
/* before actual INSERT into database                  */

typedef struct href_struct {
	char *href;
	int referrer;
	int hops;
	int stored;
} HREF;


/* Fixme! It is to be removed.                  */
/* Combine ShowStat and GetStat in one function */

typedef struct stat_struct {
	int status;
	int expired;
	int total;
} STAT;


/* Structure to store server parameters */
typedef struct server_struct {
	char *url;
	char *charset;
	char *proxy;
	char *basic_auth;
	char *htdb_list;
	char *htdb_doc;
	int  proxy_port;
	int  period;
	int  hint;
	int  net_errors;
	int  max_net_errors;
	int  read_timeout;
	int  outside;
	int  maxhops;
	int  gindex;
	int  gfollow;
	int  deletebad;
	int  userobots;
	int  bodyweight;
	int  titleweight;
	int  urlweight;
	int  descweight;
	int  keywordweight;
} SERVER;

/* All componetns in this structure correspondent */
/* to fields in the "url" table                   */

typedef struct url_struct {
	int	url_id;
	int	status;
	int	size;
	int	rating;
	int	referrer;
	int	tag;
	int	hops;
	char	*url;
	char	*content_type;
	char	*last_modified;
	char	*title;
	char 	*text;
	char	*last_index_time;
	char	*next_index_time;
	char	*keywords;
	char	*description;
	char	*crc;
} DOCUMENT;


/* Indexer */
typedef struct indexer_struct{
	int handle;		/* Indexer handle for multithreaded version */
	int mwords;		/* Number of memory allocated for words     */
	int nwords;		/* Real number of words in list             */
	int swords;		/* Number of words in sorted list           */
	UDM_WORD *Word;		/* Word list  itself                        */
	int DBError;	
	int action;		/* Callback function may request action     */
	char state[128];	/* Currently unused                         */
	char *buf;		/* Buffer for document                      */
	char *charset;		/* Charset string is stored here            */
	void * db;              /* Database handle                          */
	UDM_LANG lang[MAXLANGPERDOC]; /* language guesser variable          */
} INDEXER;
