#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_PTHREAD
#if (WIN32|WINNT)
#include <windows.h>
#else
#include <pthread.h>
#endif
#endif
#include "common.h"


#ifdef HAVE_PTHREAD
static int active_threads=0;
#if (WIN32|WINNT)
static CRITICAL_SECTION robots_mutex;
static CRITICAL_SECTION hrefs_mutex;
static CRITICAL_SECTION docs_mutex;
static CRITICAL_SECTION active_mutex;
static CRITICAL_SECTION target_mutex;
static CRITICAL_SECTION indexed_mutex;
static CRITICAL_SECTION start_mutex;
static CRITICAL_SECTION query_mutex;
#else
static pthread_mutex_t robots_mutex;
static pthread_mutex_t hrefs_mutex;
static pthread_mutex_t docs_mutex;
static pthread_mutex_t active_mutex;
static pthread_mutex_t target_mutex;
static pthread_mutex_t indexed_mutex;
static pthread_mutex_t start_mutex;
static pthread_mutex_t query_mutex;
pthread_mutex_t host_mutex;
#endif

__INDLIB__ int InitMutexes(){
#if (WIN32|WINNT)
	InitializeCriticalSection(&robots_mutex);
	InitializeCriticalSection(&hrefs_mutex);
	InitializeCriticalSection(&docs_mutex);
	InitializeCriticalSection(&active_mutex);
	InitializeCriticalSection(&target_mutex);
	InitializeCriticalSection(&indexed_mutex);
	InitializeCriticalSection(&start_mutex);
	InitializeCriticalSection(&query_mutex);
#else
	pthread_mutex_init(&robots_mutex,NULL);
	pthread_mutex_init(&hrefs_mutex,NULL);
	pthread_mutex_init(&docs_mutex,NULL);
	pthread_mutex_init(&active_mutex,NULL);
	pthread_mutex_init(&target_mutex,NULL);
	pthread_mutex_init(&indexed_mutex,NULL);
	pthread_mutex_init(&start_mutex,NULL);
	pthread_mutex_init(&host_mutex,NULL);
	pthread_mutex_init(&query_mutex,NULL);
#endif
	return(0);
}

__INDLIB__ int DestroyMutexes(){
#if (WIN32|WINNT)
	DeleteCriticalSection(&robots_mutex);
	DeleteCriticalSection(&hrefs_mutex);
	DeleteCriticalSection(&docs_mutex);
	DeleteCriticalSection(&active_mutex);
	DeleteCriticalSection(&target_mutex);
	DeleteCriticalSection(&indexed_mutex);
	DeleteCriticalSection(&start_mutex);
	DeleteCriticalSection(&query_mutex);
#else
	pthread_mutex_destroy(&robots_mutex);
	pthread_mutex_destroy(&hrefs_mutex);
	pthread_mutex_destroy(&docs_mutex);
	pthread_mutex_destroy(&active_mutex);
	pthread_mutex_destroy(&target_mutex);
	pthread_mutex_destroy(&indexed_mutex);
	pthread_mutex_destroy(&start_mutex);
	pthread_mutex_destroy(&host_mutex);
	pthread_mutex_destroy(&query_mutex);
#endif
	return(0);
}

__INDLIB__ int IncActive(){
int res;
	LOCK_ACTIVE();
	active_threads++;
	res=active_threads;
	UNLOCK_ACTIVE();
	return(res);
}
__INDLIB__ int DecActive(){
int res;
	LOCK_ACTIVE();
	active_threads--;
	res=active_threads;
	UNLOCK_ACTIVE();
	return(res);
}
__INDLIB__ int GetActive(){
int res;
	LOCK_ACTIVE();
	res=active_threads;
	UNLOCK_ACTIVE();
	return(res);
}
#endif
