#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <ctype.h>
#ifdef HAVE_PTHREAD
#endif

#include "common.h"
#include "search.h"
#include "indexer.h"
#include "charset.h"
#include "log.h"
#include "dbms.h"
#include "parseurl.h"
#include "spell.h"

#define T_UNKNOWN	0
#define T_TOP		1
#define T_BOT		2
#define T_RESTOP	3
#define T_RESBOT	4
#define T_RESULT	5
#define T_ERROR		6
#define T_NOTFOUND	7
#define T_CLONE		8
#define T_NAVIGATOR	9
#define T_NAVLEFT	10
#define T_NAVBAR0	11
#define T_NAVBAR1	12
#define T_NAVRIGHT	13



typedef struct var_struct{
	char *name;
	char *value;
} TVAR;

typedef struct template_struct{
	int where;
	char *str;
} TEMPLATE;


#define MAXTEMPLATE 1024
static TEMPLATE Template[MAXTEMPLATE];
static int ntempl=0;
static void * db;

#define MAXVARIABLE	32
TVAR Variables[MAXVARIABLE];
int nVar=0;
int Randoms[MAXRANDOM];


/* Template variables */
static char error[BUFSIZ]="";
static char wordinfo[STRSIZ]="";
static char href[STRSIZ]="";
static char fullhref[STRSIZ]="";
static char *query_form_escaped=0;
static char *crc=0;
static char self[STRSIZ]="";
static char *title=0;
static char *text=0;
static int  size=0;
static int  rating=0;
static char *descr=0;
static char *keywords=0;
static int  url_id,docnum,first=0,last=0,found=0;
static int num_pages=0, navpage=0, ps=0, pps=10,np=0,is_next=0;
static char *mode=NULL;
static char *sort=NULL;
static char *ul_unescaped=NULL;


char* htmlspecialchars(char *str){
char *pos, *p;
int len;
	if(!str)return(NULL);
	len = strlen(str)+1;
	for(pos = str; *pos; pos++){
		switch (*pos){
			case '&': len+=4; break;
			case '"': len+=5; break;
			case '<':
			case '>': len+=3; break;
		}
	}
	p = (char*) malloc(len+1);*p = 0;
	for(pos=p; *str; str++) {
	switch (*str){
		case '&': strcat(p, "&amp;"); pos+=5; break;
		case '"': strcat(p, "&quot;");pos+=6; break;
		case '<': strcat(p, "&lt;");pos+=4; break;
		case '>': strcat(p, "&gt;");pos+=4; break;
		default:
			*pos = *str;
			pos++;*pos=0;
		}
	}
	return p;
}

char * Unescape(char *d,char *s){
int hi,lo=0;
char *dd;
	if((d==NULL)||(s==NULL))return(0);
	dd=d;
	while(*s){
		if(*s=='%'){
			if(strchr("0123456789",*(++s))) hi=*s-'0';
			else hi=*s-'A'+10;
			if(strchr("0123456789",*(++s))) lo=*s-'0';
			else lo=*s-'A'+10;
			*d=hi*16+lo;
		}else
		if(*s=='+'){
			*d=' ';
		}else{
			*d=*s;
		}
		s++; d++;
	}
	*d=0;return(dd);
}

char *Escape(char *d,char *s,char what){
char *dd;
	if((d==NULL)||(s==NULL))return(0);
	dd=d;
	while(*s){
		if(*s==what){
			if(what=='"') {strcpy(d,"&quot;"); d+=5;}else
			if(what==' ') {*d='+';}
		}else{
			*d=*s;
		}
		s++; d++;
	}
	*d=0;
	return(dd);
}

char * EscapeURL(char *d,char *s){
char *dd;
	if((d==NULL)||(s==NULL))return(0);
	dd=d;
	while(*s){
		if(strchr("%&<>+[](){}/?#'\"\\;,",*s)){
			sprintf(d,"%%%x",(int)*s);
			*(d+1)=toupper(*(d+1));
			*(d+2)=toupper(*(d+2));
			d+=2;
		}else
		if(*s==' '){
			*d='+';
		}else{
			*d=*s;
		}
		s++;d++;
	}
	*d=0;
	return(dd);
}

int AddVariable(char *name,char *value){
	if(nVar<MAXVARIABLE){
		Variables[nVar].name=strdup(name);
		Variables[nVar].value=strdup(value);
		nVar++;
		return(0);
	}
	return(1);
}

int PrintTemplate(char * Target,int where,char *url,char *content_type,char *last_modified);

static int PrintOneTemplate(char * Target,char *s,int where,char *url,char *content_type,char *last_modified){
DOCUMENT *r=NULL;
int rnum,tmp;

	*Target=0;
	while(*s){
		if((*s=='\\')&&(*(s+1)=='$')){
			sprintf(STREND(Target),"$");
			s+=2;
			continue;
		}
		if(*s!='$'){
			sprintf(STREND(Target),"%c",*s);
			s++;
			continue;
		}
		
		s++;
		switch(*s){
		case 'A': sprintf(STREND(Target),"%s",self);break;
		case 'Q': sprintf(STREND(Target),"%s",query_form_escaped?query_form_escaped:"");break;
		case 'E': sprintf(STREND(Target),"%s",error);break;
		case 'W': sprintf(STREND(Target),"%s",wordinfo);break;
		case 'f': sprintf(STREND(Target),"%d",first);break;
		case 'l': sprintf(STREND(Target),"%d",last);break;
		case 't': sprintf(STREND(Target),"%d",found);break;
		case 'm': sprintf(STREND(Target),"%s",mode?mode:"all");break;
		case 's': sprintf(STREND(Target),"%s",sort?sort:"rate");break;

		case 'r': {
			s++;
			rnum=atoi(s);
			if((rnum>=0)&&(rnum<MAXRANDOM))
				sprintf(STREND(Target),"%d",Randoms[rnum]);
			while((*s)&&(isdigit(*(s+1))))
				s++;
			break;
		}
		case 'V': {
			if(num_pages>1)
				PrintTemplate(STREND(Target),T_NAVIGATOR,"","","");
			break;
		}
		case 'D':
			s++;
			switch(*s){
			case 'U':sprintf(STREND(Target),"%s",url?url:"");break;
			case 'T':sprintf(STREND(Target),"%s",title?title:"");break;
			case 'C':sprintf(STREND(Target),"%s",content_type?content_type:"");break;
			case 'M':sprintf(STREND(Target),"%s",last_modified?last_modified:"");break;
			case 'X':sprintf(STREND(Target),"%s",text?text:"");break;
			case 'R':sprintf(STREND(Target),"%d",rating);break;
			case 'S':sprintf(STREND(Target),"%d",size);break;
			case 'N':sprintf(STREND(Target),"%d",docnum);break;
			case 'D':sprintf(STREND(Target),"%s",descr);break;
			case 'K':sprintf(STREND(Target),"%s",keywords);break;
			default :sprintf(STREND(Target),"%c",*s);
			}
			break;

		case 'C':
			s++;
			switch(*s){
			case 'L':
			if((crc)&&(*crc)&&(where==T_RESULT)){
				if((r=CloneList(db,crc))){
					while(r->url_id){
						if(r->url_id!=url_id)
						PrintTemplate(STREND(Target),T_CLONE,r->url,r->content_type,r->last_modified);
						r++;
					}
				}
			}
			break;
			default : sprintf(STREND(Target),"%c",*s);
			}
			break;
		case 'N':
			s++;
			switch(*s){
				case 'L':
				if (np > 0) {
					if(np>1)sprintf(fullhref,"%s&np=%d",href,np-1);
					else strcpy(fullhref,href);
					PrintTemplate(STREND(Target),T_NAVLEFT,"","","");
				}
				break;
				case 'B':
				for(navpage = 1; navpage <= num_pages;navpage++){
					tmp = np + 1 - pps/2;
					if (tmp < 1) tmp = 1;
					if (tmp > num_pages - pps + 1)
					tmp = num_pages - pps + 1;
					if (navpage >= tmp && navpage <(tmp+pps)){
						if(navpage>1)sprintf(fullhref,"%s&np=%d",href,navpage-1);
						else strcpy(fullhref,href);
						if (np == navpage-1)
						PrintTemplate(STREND(Target),T_NAVBAR0,"","","");
					else
						PrintTemplate(STREND(Target),T_NAVBAR1,"","","");
					}
				}
			break;
				case 'R':
				if (is_next) {
					sprintf(fullhref,"%s&np=%d",href,np+1);
					PrintTemplate(STREND(Target),T_NAVRIGHT,"","","");
				}
				break;
				case 'H': sprintf(STREND(Target),"%s",fullhref);break;
				case 'P': sprintf(STREND(Target),"%d",navpage);break;
			}
			break;
		default: 
			if(!strncmp(s,"ps",2)){
				sprintf(STREND(Target),"%d",ps?ps:20);
				s++;
			}else
			if(!strncmp(s,"ul",2)){
				sprintf(STREND(Target),"%s",ul_unescaped?ul_unescaped:"");
				s++;
			}else
			sprintf(STREND(Target),"%c",*s);
		}
		s++;
	}
	return(0);
}

static int PrintOption(char * Target,int where,char * option){
TAG tag;
char *s;
int len;
char tmp[STRSIZ]="";

	if(!(s=strchr(option,'>'))){
		sprintf(STREND(Target),"%s",option);
		return(0);
	}
	len=s-option;
	s=strdup(option);
	s[len+1]=0;
	parsetag(&tag,s);
	if(tag.selected&&tag.value){
		int i;
		char *value;
		PrintOneTemplate(tmp,option,where,"","","");
		freetag(&tag);
		parsetag(&tag,tmp);
		//---
		value=tag.selected?tag.selected:"";
		if(*value=='$'){
			for(i=0;i<nVar;i++){

				if(!strcmp(Variables[i].name,"ul")&&
					!strcmp(Variables[i].value,tag.value)){
					value=Variables[i].value;
					break;
				}
			}
		}
		sprintf(STREND(Target),"<OPTION VALUE=\"%s\"%s>",tag.value,
			strcmp(tag.value,value)?"":" SELECTED");
		PrintOneTemplate(STREND(Target),option+len+1,where,"","","");
	}else{	
		sprintf(STREND(Target),"%s",option);
	}
	freetag(&tag);
	return(0);
}


int PrintTemplate(char * Target,int where,char *url,char *content_type,char *last_modified){
int i;
	*Target=0;
	for(i=0;i<ntempl;i++){
		if(Template[i].where!=where)continue;
		if(!STRNCASECMP(Template[i].str,"<OPTION")){
			PrintOption(STREND(Target),where,Template[i].str);
			continue;
		}
		PrintOneTemplate(STREND(Target),Template[i].str,where,url,content_type,last_modified);
	}
	return(0);
}


int LoadTemplate(char *name, char *query){
FILE *file;
int where=0,variables=0;
char *s,*lasttok=NULL;
char str[BUFSIZ];
char first_letters[64]="";

	file=fopen(name,"r");
	if(!file)return(1);

	while((fgets(str,sizeof(str),file))&&(ntempl<MAXTEMPLATE)){
		if(!STRNCMP(str,"<!--top-->"))		where=T_TOP;
		else
		if(!STRNCMP(str,"<!--bottom-->"))	where=T_BOT;
		else
		if(!STRNCMP(str,"<!--restop-->"))	where=T_RESTOP;
		else
		if(!STRNCMP(str,"<!--resbot-->"))	where=T_RESBOT;
		else
		if(!STRNCMP(str,"<!--res-->"))		where=T_RESULT;
		else
		if(!STRNCMP(str,"<!--notfound-->"))	where=T_NOTFOUND;
		else
		if(!STRNCMP(str,"<!--error-->"))	where=T_ERROR;
		else
		if(!STRNCMP(str,"<!--clone-->"))	where=T_CLONE;
		else
		if(!STRNCMP(str,"<!--navigator-->"))	where=T_NAVIGATOR;
		else
		if(!STRNCMP(str,"<!--navleft-->"))	where=T_NAVLEFT;
		else
		if(!STRNCMP(str,"<!--navbar1-->"))	where=T_NAVBAR1;
		else
		if(!STRNCMP(str,"<!--navbar0-->"))	where=T_NAVBAR0;
		else
		if(!STRNCMP(str,"<!--navright-->"))	where=T_NAVRIGHT;
		else
		if(!STRNCMP(str,"<!--/top-->"))		where=0;
		else
		if(!STRNCMP(str,"<!--/bottom-->"))	where=0;
		else
		if(!STRNCMP(str,"<!--/restop-->"))	where=0;
		else
		if(!STRNCMP(str,"<!--/resbot-->"))	where=0;
		else
		if(!STRNCMP(str,"<!--/res-->"))		where=0;
		else
		if(!STRNCMP(str,"<!--/notfound-->"))	where=0;
		else
		if(!STRNCMP(str,"<!--/error-->"))	where=0;
		else
		if(!STRNCMP(str,"<!--/clone-->"))	where=0;
		else
		if(!STRNCMP(str,"<!--variables")){	where=0;variables=1;}
		else
		if(!STRNCMP(str,"<!--/navigator-->"))	where=0;
		else
		if(!STRNCMP(str,"<!--/navleft-->"))	where=0;
		else
		if(!STRNCMP(str,"<!--/navbar1-->"))	where=0;
		else
		if(!STRNCMP(str,"<!--/navbar0-->"))	where=0;
		else
		if(!STRNCMP(str,"<!--/navright-->"))	where=0;
		else
		if(!STRNCMP(str,"-->")){	
			if(variables){
				where=0;
				variables=0;
			}
		}else{
			if(where){
				Template[ntempl].where=where;
				Template[ntempl].str=strdup(str);
				ntempl++;
			}
			if((variables)&&(s=gettoken(str,"\r\n",&lasttok))){
				if(!STRNCASECMP(str,"DBHost"))
					DBHost=strdup(trim_chars(str+6,"= \t\r\n"));
				else
				if(!STRNCASECMP(str,"DBName"))
					DBName=strdup(trim_chars(str+6,"= \t\r\n"));
				else
				if(!STRNCASECMP(str,"DBUser"))
					DBUser=strdup(trim_chars(str+6,"= \t\r\n"));
				else
				if(!STRNCASECMP(str,"DBPass"))
					DBPass=strdup(trim_chars(str+6,"= \t\r\n"));
				else
				if(!STRNCASECMP(str,"DBPort"))
					DBPort=atoi(trim_chars(str+6,"= \t\r\n"));
				else
				if(!STRNCASECMP(str,"ResultsPerPage"))
					ps=atoi(trim_chars(str+5,"= \t\r\n"));
				else
				if(!STRNCASECMP(str,"PagesPerScreen"))
					pps=atoi(trim_chars(str+5,"= \t\r\n"));
				else
				if(!STRNCASECMP(str,"LocalCharset")){
					LCindex=getcharset(trim_chars(str+12,"= \r\t\n"));
					makeseparator(SEPARATOR,LCindex);
					if(query){
						char *lasttok=NULL;
						char *w;
						tolower_string(query,LCindex);
						strcpy(str,query);
						w=gettoken(str,SEPARATOR,&lasttok);
						while(w){
							int len;
							len=strlen(first_letters);
							first_letters[len]=w[0];
							first_letters[len+1]=0;
							w=gettoken(NULL,SEPARATOR,&lasttok);
						}
					}
				}else
				if(!STRNCASECMP(str,"R")){
					int i,j;
					float frand;
					i=atoi(str+1);
					j=atoi(trim_chars(str+1,"= \t\r\n"));
					if((i>0)&&(i<MAXRANDOM)){
						frand=rand();
						frand=frand/RAND_MAX*j;
						Randoms[i]=frand;
					}
				}else
				if(!STRNCASECMP(str,"Affix")){
					char lang[20]="";char str1[STRSIZ]="";
					if(query)
					if(sscanf(str+5,"%s%s",lang,str1)){
						if(*str1=='/')strcpy(str,str1);
						else	sprintf(str,"%s/%s",CONFIGDIR,str1);
						ImportAffixes(lang,str);
					}
				}else
				if(!STRNCASECMP(str,"Spell")){
				char lang[20]=""; char str1[STRSIZ]="";
					if(query)
					if(sscanf(str+5,"%s%s",lang,str1)){
						if(*str1=='/')strcpy(str,str1);
						else	sprintf(str,"%s/%s",CONFIGDIR,str1);
						ImportDictionary(lang,str,1,first_letters);
					}
				}
			}
		}
	}
	fclose(file);
	if(query){
		if(!LCindex){
			makeseparator(SEPARATOR,LCindex);
			tolower_string(query,LCindex);
		}
		SortDictionary();
	}
	return(0);
}


int main(int argc, char **argv) {
time_t tclock;
char str[STRSIZ]="";
char str1[STRSIZ]="";
char template[STRSIZ]="";
char query_string[STRSIZ]="";
char ul_str[STRSIZ]="";
char *q=NULL,*ul=NULL,*tagstr=NULL,*sq=NULL;
char *env,*s;
int i,sort_order=ORD_RATE;
int search_mode=MOD_ALL;
char *lasttok=NULL;

DOCUMENT * r=NULL, * r1=NULL;
char Target[STRSIZ*16]="";

	
	/* Determine self and template name */
	if((env=getenv("QUERY_STRING"))){
		strcpy(query_string,env);
		printf("Content-type: text/html\r\n\r\n");
		if((env=getenv("REDIRECT_STATUS"))){
			env=getenv("REDIRECT_URL");
			strcpy(self,env?env:"search.cgi");
			env=getenv("PATH_TRANSLATED");
			strcpy(template,env?env:"");
		}else{
			env=getenv("SCRIPT_NAME");
			strcpy(self,env?env:"search.cgi");

			if(strcmp(CONFIGDIR,".")){
				sprintf(template,"%s/%s", CONFIGDIR,(s=strrchr(self,SLASH))?(s+1):(self));
			}else{
				env=getenv("SCRIPT_FILENAME");
				strcpy(template,env?env:"search.cgi");
			}
			if((s=strrchr(template,'.'))){
				*s=0;strcat(template,".htm");
			}else{
				strcpy(str,"search.htm");
			}
		}
	}else{
		if(argv[1])sprintf(query_string,"q=%s",argv[1]);
		sprintf(template,"%s/%s", CONFIGDIR,"search.htm");
	}


	/* Parse Query String */
	s=gettoken(query_string,"&",&lasttok);
	while(s){
		if(!STRNCMP(s,"q=")){
			q=strdup(Unescape(str,s+2));
			sq=strdup(EscapeURL(str,q));
			query_form_escaped=strdup(Escape(str,q,'"'));
		}else
		if(!STRNCMP(s,"np="))np=atoi(s+3);
		else
		if(!STRNCMP(s,"ps="))ps=atoi(s+3);
		else
		if(!STRNCMP(s,"ul=")){
#ifdef USE_AUTO_WILD
			sprintf(str1,"%%%s%%",Unescape(str,s+3));
#else
			sprintf(str1,"%s",Unescape(str,s+3));
#endif
			AddVariable("ul",str);
			free_it(ul_unescaped);
			ul_unescaped=strdup(str);
			AddURLLimit(str1);
			ul=strdup(s+3);
			sprintf(STREND(ul_str),"&ul=%s",ul);
		}else
		if((!STRNCMP(s,"t="))&&(*(s+2))){
			AddTagLimit(atoi(s+2));
		}else
		if(!STRNCMP(s,"m=")){
			mode=s+2;
			if(!strcmp(mode,"all"))search_mode=MOD_ALL;
			if(!strcmp(mode,"any"))search_mode=MOD_ANY;
		}else
		if(!STRNCMP(s,"s=")){
			sort=s+2;
			if(!strcmp(sort,"date"))sort_order=ORD_DATE;
			if(!strcmp(sort,"rate"))sort_order=ORD_RATE;
		}
		s=gettoken(NULL,"&",&lasttok);
	}

	tclock=time(0);srand(tclock);
	for(i=0;i<MAXRANDOM;i++)Randoms[i]=0;

	InitCharset();
	db=AllocDB();

	if(LoadTemplate(template,q)){
		printf("<html><body>Can't open template file '%s'!</body></html>\n",template);
		return(0);
	}

	PrintTemplate(Target,T_TOP,"","","");printf("%s",Target);

	if(!q){
		PrintTemplate(Target,T_BOT,"","","");printf("%s",Target);
		exit(0);
	}


#ifdef HAVE_PTHREAD
	InitMutexes();
#endif

	if((r=Find(db,q,np,ps?ps:20,search_mode,sort_order,wordinfo,&found))){
		last=first=docnum=(ps?ps:20)*np+1;
		r1=r;while((++r1)->url_id)last++;
		num_pages = found/(ps?ps:20);
		is_next=last<found;
		if((ps?ps:20)*num_pages<found)num_pages++;

		PrintTemplate(Target,T_RESTOP,"","","");printf("%s",Target);
		sprintf(href,"%s?q=%s",self,sq);

		if(tagstr)sprintf(STREND(href),"&t=%s",tagstr);
		//if(ul)sprintf(STREND(href),"&ul=%s",ul);
		strcpy(STREND(href),ul_str);
		if(mode)sprintf(STREND(href),"&m=%s",mode);
		if(sort)sprintf(STREND(href),"&s=%s",sort);
		if(ps)sprintf(STREND(href),"&ps=%d",ps);

		while(r->url_id){
			url_id=r->url_id;
			title=htmlspecialchars(r->title);
			text=htmlspecialchars(r->text);
			size=r->size;
			rating=r->rating;
			descr=htmlspecialchars(r->description);
			keywords=htmlspecialchars(r->keywords);
			crc=r->crc;
			PrintTemplate(Target,T_RESULT,r->url,r->content_type,r->last_modified);
			printf("%s",Target);
			free_it(title);free_it(text);
			free_it(descr);free_it(keywords);
			r++;docnum++;
		}
		PrintTemplate(Target,T_RESBOT,"","","");printf("%s",Target);
	}else{
		if(DBErrorCode(db)){
			strcpy(error,DBErrorMsg(db));
			PrintTemplate(Target,T_ERROR,"","","");printf("%s",Target);
		}else{
			PrintTemplate(Target,T_NOTFOUND,"","","");printf("%s",Target);
		}
	}
	PrintTemplate(Target,T_BOT,"","","");printf("%s",Target);
	FreeDB(db);

#ifdef HAVE_PTHREAD
	DestroyMutexes();
#endif
	return(0);
}
