#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "config.h"
#include "common.h"
#include "spell.h"
#include "robots.h"
#if (WIN32|WINNT)
#include <windows.h>
#include <windef.h>
#include <time.h>
#else
#include <unistd.h>
#include <sys/time.h>
#endif
#include "log.h"
#include "dbms.h"
#include "charset.h"
#if (HAVE_MYSQL)
#include "mysql.h"
#define DBVER "MySQL"
#define DB_DEFAULT	DB_MYSQL
#define HAVE_SQL_LIMIT
#define HAVE_SQL_IN
#elif (HAVE_PGSQL)
#include "libpq-fe.h"
#define DBVER "PgSQL"
#define DB_DEFAULT	DB_PGSQL
#define HAVE_SQL_LIMIT
#define HAVE_SQL_IN
#elif HAVE_MSQL
#define DBVER "mSQL"
#define DB_DEFAULT	DB_MSQL
#include "msql.h"
#elif HAVE_IODBC
#define DBVER "iODBC"
#define DB_DEFAULT	DB_UNK
#include "isql.h"
#include "isqlext.h"
#elif HAVE_VIRT
#define DBVER "Virtuoso"
#define DB_DEFAULT	DB_VIRT
#include "iodbc.h"
#include "isql.h"
#include "isqlext.h"
#elif HAVE_UNIXODBC
#define DBVER "unixODBC"
#define DB_DEFAULT	DB_UNK
#include "sql.h"
#include "sqlext.h"
#elif HAVE_SOLID
#define DBVER "Solid"
#define DB_DEFAULT	DB_SOLID
#include "cli0cli.h"
#elif HAVE_IBASE
#include "ibase.h"
#define DBVER "InterBase"
#define DB_DEFAULT	DB_IBASE
#else
#error Dabase was not selected. You shoud choose database using configure --with-your_database. Take a look at INSTALL file
#endif

#ifdef DEBUG_SQL
#include <sys/times.h>
static unsigned long start_timer(void){
#ifdef WIN32
	return clock();
#else
	struct tms tms_tmp;
	return times(&tms_tmp);
#endif
}
#endif


#ifdef USE_MULTIDICT
#define NDICTS	18
#define MINDICT	2
#define MAXDICT	NDICTS
static int dictlen[NDICTS]={2,2,2,3,4,5,6,7,8,9,10,11,12,16,16,16,16,32};
#define	DICTNUM(l)	((l>16)?dictlen[17]:dictlen[l])
#else
#define DICTNUM(l)	0
#endif

#define	URL_DELETE_CACHE 50
#define	URL_SELECT_CACHE 32

char DBVersion[]   = DBVER;
char *DBHost       = NULL;
char *DBName       = NULL;
char *DBUser       = NULL;
char *DBPass       = NULL;
int  DBType        = DB_DEFAULT;
int  DBPort        = 0;
int  GMToffs       = 0;
int  CurURLNumber  = 0;
int  have_targets  = 1;
int  index_started = 0;
int  mark_done     = 1;

static  int  MaxURLNumber=-1;
static  char tagstr[STRSIZ]="";
static  char statusstr[BUFSIZ]="";
static  char urlstr[BUFSIZ]="";
static	STOP_WORD *stoplist=NULL;
static  int nstoplist=0;
static	int currow=0;
static	int nrows=0;

static time_t now(){ /* !!! is it reentrant? */
time_t tloc;
	return(time(&tloc));
}

#ifdef HAVE_MYSQL
char * escstr(char *x,char *y){
	mysql_escape_string(x,y,strlen(y));
	return(x);
}
#else
char * escstr(char *x,char *y){
char *s;
	s=x;
	while(*y){
		switch(*y){
			case '\'':
			case '\\':
				*x='\\';x++;
			default:*x=*y;
		}
		x++;y++;
	}
	*x=0;return(s);
}
#endif


/************************ MYSQL **************************************/
#if   (HAVE_MYSQL)
typedef struct struct_db {
	MYSQL mysql;
	MYSQL_RES *res;
	MYSQL_RES *res1;
	int connected;
	int res_limit;
	int errcode;
	char errstr[STRSIZ];
} DB;
typedef MYSQL_RES UDB_RES;
#define SQL_NUM_ROWS(x)	mysql_num_rows(x)
#define SQL_FREE(x)	mysql_free_result(x)
#define lock_dict(db)	sql_query(db,"LOCK TABLES dict WRITE")
#define flush_dict(db)	sql_query(db,"UNLOCK TABLES");sql_query(db,"LOCK TABLES dict WRITE")
#define unlock_dict(db)	sql_query(db,"UNLOCK TABLES")
#define lock_url(db)	sql_query(db,"LOCK TABLES url WRITE")
#define unlock_url(db)	sql_query(db,"UNLOCK TABLES")

static int InitDB(DB*db)
{
	if (!(mysql_connect(&(db->mysql),DBHost,DBUser,DBPass))){
		db->errcode=1;
		return(1);
	}
	if ((mysql_select_db(&(db->mysql),DBName?DBName:"udmsearch"))){
		db->errcode=1;
		return(1);
	}
	db->connected=1;
	return(0);
}
static char * sql_value(MYSQL_RES *res,int i,int j){
MYSQL_ROW row;
	if(i>=mysql_num_rows(res))return(0);		
	mysql_data_seek(res,i);
	row=mysql_fetch_row(res);
	if(row)return(row[j]?row[j]:"");
	else	return(0);
}
static void CloseDB(DB *db){
	if(db->connected)mysql_close(&(db->mysql));
	db->connected=0;
}
void displayError(DB * db){
	sprintf(db->errstr,"#%d: %s",
		mysql_errno(&((DB*)db)->mysql),
		mysql_error(&((DB*)db)->mysql));
}
static int safe_mysql_query(DB * db,char *query){
int err,i;
	if(!db->connected){
		InitDB(db);
		if(db->errcode)
			return(db->errcode);
	}
	for(i=0;i<2;i++){
		if((err=mysql_query(&db->mysql,query))){
			if((mysql_errno(&db->mysql)==CR_SERVER_LOST)||
			(mysql_errno(&db->mysql)==CR_SERVER_GONE_ERROR)||
			(mysql_errno(&db->mysql)==ER_SERVER_SHUTDOWN)){
				UDMSLEEP(5);
			}else
				return(err);
		}else
			return(err);
	}
	return(err);
}
static MYSQL_RES * sql_query(DB * db,char *query){
int res;
#ifdef DEBUG_SQL
double t;
unsigned long ticks;
	ticks=start_timer();
#endif
	db->errcode=0;
	res=safe_mysql_query(db,query);
#ifdef DEBUG_SQL
	ticks=start_timer()-ticks;
	t=(float)ticks/CLOCKS_PER_SEC;
	printf("<P>SQL <b>%.2f</b> sec: %s<HR>\n",t,query);
#endif
	if(res){
		displayError(db);
		if(mysql_errno(&(db->mysql))!=ER_DUP_ENTRY){
			db->errcode=1;
		}
		return(0);
	}
	return(mysql_store_result(&db->mysql));
}
/*********************************** POSTGRESQL *********************/
#elif (HAVE_PGSQL)
typedef struct struct_db {
	PGconn     *pgsql;
	PGresult   *res;
	PGresult   *res1;
	int res_limit;
	int connected;
	int errcode;
	char errstr[STRSIZ];
} DB;
typedef PGresult UDB_RES;
static PGresult   *urlres=NULL;
#define SQL_NUM_ROWS(x)	PQntuples(x)
#define SQL_FREE(x)	PQclear(x)
#define sql_value(x,y,z) PQgetvalue(x,y,z)
#define lock_dict(db)	sql_query(db,"BEGIN WORK")
#define	flush_dict(db)	sql_query(db,"END WORK");sql_query(db,"BEGIN WORK")
#define unlock_dict(db)	sql_query(db,"END WORK")
#define lock_url(db)	sql_query(db,"BEGIN WORK");sql_query(db,"LOCK url")
#define unlock_url(db)	sql_query(db,"END WORK")
static int InitDB(DB *db){
char port[8];
	sprintf(port,"%d",DBPort);
	db->pgsql = PQsetdbLogin(DBHost, DBPort?port:0, 0, 0, DBName, DBUser, DBPass);
	if (PQstatus(db->pgsql) == CONNECTION_BAD){
		db->errcode=1;
		return(1);
	}
	db->connected=1;
	return(0);
}
static void CloseDB(DB *db){
	if(db->connected)PQfinish(db->pgsql);
	db->connected=0;
}
void displayError(DB * db){
	sprintf(db->errstr, "%s", PQerrorMessage(db->pgsql));
}
static PGresult * safe_pgsql_query(DB *db,char *q){
	if(!(db->connected)){
		InitDB(db);
		if(db->errcode)return(0);
	}
	return(PQexec(db->pgsql,q));
}
static PGresult * sql_query(void *db,char *q){
PGresult * res;
#ifdef DEBUG_SQL
double t;
unsigned long ticks;
	ticks=start_timer();
#endif

	((DB*)db)->errcode=0;
	res=safe_pgsql_query((DB*)db,q);
	if(!res){
		displayError(db);
		((DB*)db)->errcode=1;
		return(0);
	}
	if((PQresultStatus(res)!=PGRES_COMMAND_OK)&&(PQresultStatus(res)!=PGRES_TUPLES_OK)){
		displayError(db);
		if(!strstr(PQerrorMessage(((DB*)db)->pgsql),"duplicate"))
			((DB*)db)->errcode=1;
		return(0);
	}
#ifdef DEBUG_SQL
	ticks=start_timer()-ticks;
	t=(float)ticks/CLOCKS_PER_SEC;
	printf("<P>SQL <b>%.2f</b> sec: %s<HR>\n",t,q);
#endif
	return(res);
}
/*****************************  miniSQL *******************************/
#elif (HAVE_MSQL)
typedef struct struct_db {
	int		msql;
	m_result	*res;
	int res_limit;
	int connected;
	int errcode;
	char errstr[STRSIZ];
} DB;
typedef m_result UDB_RES;
static m_result		*urlres=NULL;
#define SQL_NUM_ROWS(x)	msqlNumRows(x)
#define SQL_FREE(x)	msqlFreeResult(x)
#define lock_dict(db)	
#define	flush_dict(db)	
#define unlock_dict(db)	
#define lock_url(db)	
#define unlock_url(db)

static char * sql_value(m_result *res,int i,int j){
m_row row;

	if(i>=msqlNumRows(res))return("");
	msqlDataSeek(res,i);
	row=msqlFetchRow(res);
	if(row)return(row[j]?row[j]:"");
	else	return(0);
}
static int InitDB(DB *db){
char *host;
	if((host=DBHost))if(!strcmp(DBHost,"localhost"))host=NULL;
	if((db->msql=msqlConnect(host))<0){
		db->errcode=1;
		return(1);
	}
	if(msqlSelectDB(db->msql,DBName)==-1){
		db->errcode=1;
		return(1);
	}
	((DB*)db)->connected=1;
	return(0);
}
void CloseDB(DB *db){
	if(db->connected)msqlClose(db->msql);
	db->connected=0;
}
void displayError(DB *db){
	sprintf(db->errstr,"%s",msqlErrMsg);
}
static m_result * sql_query(void *db,char *query);
static int safe_msql_query(DB *db,char *query){
int err;

	if(!(db->connected)){
		InitDB(db);
		if(db->errcode)
			return(db->errcode);
	}
	if((err=msqlQuery(db->msql,query))==-1){
		if(!strstr(msqlErrMsg,"Non unique")){
			db->errcode=1;
			return(err);
		}
		return(0);
	}
	return(0);
}
static m_result * sql_query(void *db,char *query){
int err;
	((DB*)db)->errcode=0;
	if((err=safe_msql_query((DB*)db,query))){
		displayError(db);
		return(0);
	}
	return(msqlStoreResult());
}
/************************** ODBC ***********************************/
#elif (HAVE_IODBC || HAVE_UNIXODBC || HAVE_SOLID || HAVE_VIRT)
typedef struct or_struct{
	int nRows;
	int nCols;
	int freeme;
	char **items;
} iODBC_RES;
typedef iODBC_RES UDB_RES;
typedef struct struct_db {
	HDBC	hDbc;
	HENV	hEnv;
	HSTMT	hstmt;
	iODBC_RES * res;
	int res_limit;
	int connected;
	int errcode;
	char errstr[STRSIZ];
} DB;

#define SQL_NUM_ROWS(x)	(x?x->nRows:0)
#define SQL_FREE(x)	iodbc_free_result(x)
#define sql_val(x,y,z)	x->items[y*x->nCols+z]
#define sql_value(x,y,z) x?(sql_val(x,y,z)?sql_val(x,y,z):""):""
#define lock_dict(db)
#define	flush_dict(db)	
#define unlock_dict(db)	
#define lock_url(db)
#define unlock_url(db)
#define SQL_OK(rc)	((rc==SQL_SUCCESS)||(rc==SQL_SUCCESS_WITH_INFO))

static int iodbc_free_result(iODBC_RES *res){
int i;
	if(res){
		if(res->freeme){
			if(res->items){
				for(i=0;i<res->nCols*res->nRows;i++)
					if(res->items[i])
						free(res->items[i]);
				free(res->items);
			}
			free(res);
		}
	}
	return(0);
}
static int displayError(DB * db){
	UCHAR szSqlState[6];
	UCHAR szErrMsg[256];
	SDWORD naterr;
	SWORD length;
	
	db->errstr[0]=0;
	while (SQL_SUCCESS == SQLError(db->hEnv,db->hDbc,db->hstmt,szSqlState,&naterr,szErrMsg,sizeof(szErrMsg),&length))
		strcat(db->errstr, szErrMsg);
	return(0);
}
iODBC_RES * execDB(DB*db,char* sqlstr){
	RETCODE rc;
	SWORD iResColumns;
	SDWORD iRowCount;
	int i,res_count;
	char* p;
	UCHAR szColName[32];
	SWORD pcbColName;
	SWORD pfSQLType;
	UDWORD pcbColDef;
	SWORD pibScale;
	SWORD pfNullable;
	iODBC_RES *result;
	SDWORD	pcbValue;
	static char	bindbuf[(int)(32L * 1024L - 16L)];

	/* -------- 
	p=sqlstr;
	while(*p){
		if(*p=='?')*p='.';
		p++;
	}
	*/

	rc=SQLAllocStmt(db->hDbc, &(db->hstmt));
	if (!SQL_OK(rc)){
		db->errcode=1;
		return(NULL);
	}
	rc=SQLExecDirect(db->hstmt,(UCHAR*)sqlstr, SQL_NTS);
	if (!SQL_OK(rc)){
		db->errcode=1;
		return(NULL); 
	}
	rc=SQLNumResultCols(db->hstmt, &iResColumns);
	if(!SQL_OK(rc)){
		db->errcode=1;
		return(NULL);
	}
	if(!iResColumns) {
		rc=SQLRowCount(db->hstmt, &iRowCount);
		if (!SQL_OK(rc)){
			db->errcode=1;
			return(NULL);
		}
		result=NULL;
	}else{
		result=(iODBC_RES*)malloc(sizeof(iODBC_RES));
		result->nRows = 0;
		result->nCols = iResColumns;
		result->items=NULL;
		result->freeme=1;

		rc = SQL_NO_DATA_FOUND;
		for (res_count=0;(db->res_limit?(res_count<db->res_limit):1);res_count++){
			rc=SQLFetch(db->hstmt);
			if (!SQL_OK(rc)) {
				if (rc!=SQL_NO_DATA_FOUND){
					db->errcode=1;
				}
				break;
			}
			if(!result->nRows){
				result->items=(char **)malloc(1*iResColumns*sizeof(char*));
			}else{
				result->items=(char **)realloc(result->items,((result->nRows+1)*iResColumns*sizeof(char*)));
			}
			for (i = 0; i < iResColumns; i++) {
				SQLDescribeCol(db->hstmt, i+1, szColName, sizeof(szColName),
					&pcbColName, &pfSQLType, &pcbColDef,&pibScale, &pfNullable);
				SQLGetData(db->hstmt,i+1,SQL_C_CHAR,bindbuf,sizeof(bindbuf),&pcbValue);
				if (pcbValue==SQL_NULL_DATA) p = NULL;
				else p = bindbuf;
				if(p)rtrim_chars(p," ");
				sql_val(result,result->nRows,i)=strdup(p?p:"");
			}
			result->nRows++;
		}
	}
	db->res_limit=0;
	SQLFreeStmt(db->hstmt, SQL_DROP);
	db->errcode=0;
	return(result);
}
static int InitDB(DB*db){
char DSN[STRSIZ]="";
#if (HAVE_SOLID)
	sprintf(DSN,"tcp %s %d",DBHost?DBHost:"localhost",DBPort?DBPort:1313);
#else
	strcpy(DSN,DBName?DBName:"");
#endif
	db->errcode = SQLAllocEnv( &(db->hEnv) );
	if( SQL_SUCCESS != db->errcode )return -2;
	db->errcode = SQLAllocConnect( db->hEnv, &(db->hDbc) );
	if( SQL_SUCCESS != db->errcode )return -3;
	db->errcode = SQLSetConnectOption( db->hDbc, SQL_AUTOCOMMIT, SQL_AUTOCOMMIT_ON);
	if( SQL_SUCCESS != db->errcode )return -4;
	db->errcode = SQLConnect( db->hDbc, DSN, SQL_NTS, DBUser, SQL_NTS, DBPass, SQL_NTS);
	if( !SQL_OK(db->errcode)) return(-5);
	else db->errcode=0;
	db->connected=1;
	return 0;
}
void CloseDB(DB * db){
	if(db->connected){
		db->connected=0;
		db->errcode = SQLTransact( db->hEnv, db->hDbc, SQL_COMMIT);
		if( SQL_SUCCESS != db->errcode )return;
		db->errcode = SQLDisconnect( db->hDbc );
		if( SQL_SUCCESS != db->errcode )return;
		db->errcode = SQLFreeConnect( db->hDbc );
		if( SQL_SUCCESS != db->errcode )return;
		else	db->hDbc = SQL_NULL_HDBC;
		db->errcode = SQLFreeEnv( db->hEnv );
		if( SQL_SUCCESS != db->errcode )return;
		else	db->hEnv = SQL_NULL_HENV;
	}
}
static iODBC_RES * sql_query(DB * db,char *qbuf){
	iODBC_RES * res;
	
	LOCK_QUERY();
	if(!db->connected){
		InitDB(db);
		if(db->errcode){
			displayError(db);
			UNLOCK_QUERY();
			return(0);
		}else{
			db->connected=1;
		}
	}
	res=execDB(db,qbuf);
	if((db->errcode)){
		displayError(db);
		if(strstr(db->errstr,"uplicat")){ /* PgSQL,MySQL*/
			db->errcode=0;
		}else
		if(strstr(db->errstr,"nique")){ /* Solid, Virtuoso */
			db->errcode=0;
		}else{
			db->errcode=1;
			res=NULL;
		}
		SQLFreeStmt(db->hstmt, SQL_DROP);
	}
	UNLOCK_QUERY();
	return(res);
}
#elif HAVE_IBASE

typedef struct or_struct{
	int nRows;
	int nCols;
	int freeme;
	char **items;
} iBASE_RES;
typedef iBASE_RES UDB_RES;

#define SQL_NUM_ROWS(x)	(x->nRows)
#define SQL_FREE(x)	ibase_free_result(x)
#define sql_val(x,y,z) x->items[y*x->nCols+z]
#define sql_value(x,y,z) sql_val(x,y,z)?sql_val(x,y,z):""
#define lock_dict(x)
#define	flush_dict(x)	
#define unlock_dict(x)	
#define lock_url(x)
#define unlock_url(x)

#define SQL_VARCHAR(len) struct {short vary_length; char vary_string[(len)+1];}
typedef struct ibase_db {
	int		connected;
	int		errcode;
	isc_db_handle	DBH;		/* database handle */
	ISC_STATUS	status[20];	/* status vector */
	char		errstr[STRSIZ]; /* error buffer */
	UDB_RES		* res;
	int		res_limit;
} DB;

void displayError(DB * db){
char *s;
ISC_STATUS * stat;
	stat=(db->status);
	s=db->errstr;
	while(isc_interprete(s ,&stat)){
		strcat(db->errstr," ");
		s=db->errstr+strlen(db->errstr);
	}
}

int ibase_free_result(iBASE_RES *res){
int i;
	if(res){
		if(res->freeme){
			if(res->items){
				for(i=0;i<res->nCols*res->nRows;i++)
					if(res->items[i])
						free(res->items[i]);
				free(res->items);
			}
			free(res);
		}
	}
	return(0);
}
static int InitDB(DB * db){
	
	char dpb_buffer[256], *dpb, *p;
	int dpb_length, len;

	dpb = dpb_buffer;

	*dpb++ = isc_dpb_version1;

	if (DBUser != NULL && (len = strlen(DBUser))) {
		*dpb++ = isc_dpb_user_name;
		*dpb++ = len;
		for (p = DBUser; *p;) {
			*dpb++ = *p++;
		}
	}

	if (DBPass != NULL && (len = strlen(DBPass))) {
		*dpb++ = isc_dpb_password;
		*dpb++ = strlen(DBPass);
		for (p = DBPass; *p;) {
			*dpb++ = *p++;
		}
	}
	/*
	if (charset != NULL && (len = strlen(charset))) {
		*dpb++ = isc_dpb_lc_ctype;
		*dpb++ = strlen(charset);
		for (p = charset; *p;) {
			*dpb++ = *p++;
		}
	}
#ifdef isc_dpb_sql_role_name
	if (role != NULL && (len = strlen(role))) {
		*dpb++ = isc_dpb_sql_role_name;
		*dpb++ = strlen(role);
		for (p = role; *p;) {
			*dpb++ = *p++;
		}
	}
#endif
	*/

	dpb_length = dpb - dpb_buffer;

	if(isc_attach_database(db->status, strlen(DBName), DBName, &(db->DBH), dpb_length, dpb_buffer)){
		db->errcode=1;
		return(1);
	}
	/*
	if (isc_attach_database(db->status, 0, DBName, &(db->DBH), 0, NULL)){
		db->errcode=1;
		return(1);
	}
	*/
	return(0);
}
static void CloseDB(DB*db){
	if(db->connected){
		if (isc_detach_database(db->status, &(db->DBH))){
			db->errcode=1;
		}
	}
}

static iBASE_RES empty_res;
static iBASE_RES * sql_ibase_query(DB * db, char *query){
	int i, coltype;
	static char query_info[] = { isc_info_sql_stmt_type };
	char info_buffer[18];
	short l;
	long query_type;
	XSQLDA *osqlda=NULL;
	isc_tr_handle tr_handle = NULL;
	isc_stmt_handle query_handle = NULL;
	long fetch_stat;
	XSQLVAR *var;
	iBASE_RES *result=NULL;

	if(!db->connected){
		InitDB(db);
		if(db->errcode){
			displayError(db);
			db->errcode=1;
			return NULL;
		}else{
			db->connected=1;
		}
	}
	if (isc_start_transaction(db->status, &tr_handle, 1, &(db->DBH), 0, NULL)){
		db->errcode=1;
		return(0);
	}
	if (isc_dsql_allocate_statement(db->status, &(db->DBH), &query_handle)){
		db->errcode=1; 
		return(0);
	}
	if (isc_dsql_prepare(db->status, &tr_handle, &query_handle, 0, query, 1, osqlda)){
		db->errcode=1; 
		return(0);
	}
	if (!isc_dsql_sql_info(db->status, &query_handle, sizeof(query_info), query_info, sizeof(info_buffer), info_buffer)) {
		l = (short) isc_vax_integer((char ISC_FAR *) info_buffer + 1, 2);
		query_type = isc_vax_integer((char ISC_FAR *) info_buffer + 3, l);
	}

	/* Find out what kind of query is to be executed */
	if (query_type == isc_info_sql_stmt_select || query_type == isc_info_sql_stmt_select_for_upd) {
		/*
		 * Select, need to allocate output sqlda and and prepare it for use.
		 */
		osqlda = (XSQLDA *) malloc(XSQLDA_LENGTH(0));
		osqlda->sqln = 0;
		osqlda->version = SQLDA_VERSION1;

		if (isc_dsql_describe(db->status, &query_handle, 1, osqlda)) {
			free(osqlda);
			db->errcode=1;
			return NULL;
		}

		if (osqlda->sqld) {
			osqlda = (XSQLDA *) realloc(osqlda, XSQLDA_LENGTH(osqlda->sqld));
			osqlda->sqln = osqlda->sqld;
			osqlda->version = SQLDA_VERSION1;
			if (isc_dsql_describe(db->status, &query_handle, 1, osqlda)) {
				free(osqlda);
				db->errcode=1;
				return NULL;
			}
		}
		for (i = 0; i < osqlda->sqld; i++) {
			osqlda->sqlvar[i].sqlind = (short *) malloc(sizeof(short));
			coltype = osqlda->sqlvar[i].sqltype & ~1;
			switch(coltype)
				{
				case SQL_TEXT:
					osqlda->sqlvar[i].sqldata = (char *) malloc(sizeof(char)*(osqlda->sqlvar[i].sqllen));
					break;
				case SQL_VARYING:
					osqlda->sqlvar[i].sqldata = (char *) malloc(sizeof(char)*(osqlda->sqlvar[i].sqllen+2));
					break;
				case SQL_SHORT:
					osqlda->sqlvar[i].sqldata = (char *) malloc(sizeof(short));
					break;
				case SQL_LONG:
					osqlda->sqlvar[i].sqldata = (char *) malloc(sizeof(long));
					break;
				case SQL_FLOAT:
					osqlda->sqlvar[i].sqldata = (char *) malloc(sizeof(float));
					break;
				case SQL_DOUBLE:
					osqlda->sqlvar[i].sqldata = (char *) malloc(sizeof(double));
					break;
				case SQL_DATE:
				case SQL_BLOB:
				case SQL_ARRAY:
					osqlda->sqlvar[i].sqldata = (char *) malloc(sizeof(ISC_QUAD));
					break;
				}
		}
		if (isc_dsql_execute(db->status, &tr_handle, &query_handle, 1, NULL)) {
			free(osqlda);
			db->errcode=1;
			return NULL;
		}
		result=(iBASE_RES*)malloc(sizeof(iBASE_RES));
		result->nRows = 0;
		result->nCols = osqlda->sqld;
		result->items=NULL;
		result->freeme=1;

		while ((fetch_stat = isc_dsql_fetch(db->status, &query_handle, 1, osqlda)) == 0){
			int i;char buf[STRSIZ]="";
			var=osqlda->sqlvar;

			if(!result->nRows){
				result->items=(char **)malloc(1*osqlda->sqld*sizeof(char*));
			}else{
				result->items=(char **)realloc(result->items,((result->nRows+1)*osqlda->sqld*sizeof(char*)));
			}
			for(i=0;i<osqlda->sqld; i++, var++){
				if(*var->sqlind==-1)
					 buf[0]=0;  /* NULL data */
				else
				switch(var->sqltype & ~1){
					case SQL_TEXT:
					strncpy(buf,(char*)var->sqldata,var->sqllen);
					buf[var->sqllen]=0;
					break;
					case SQL_LONG:
					sprintf(buf,"%ld",*(long*)(var->sqldata));
					break;
					case SQL_SHORT:
					sprintf(buf,"%d",*(short*)(var->sqldata));
					break;
					default:
					sprintf(buf,"Unknown SQL type\n");
					buf[0]=0;
					break; 
				}
				rtrim_chars(buf," ");
				sql_val(result,result->nRows,i)=strdup(buf);
			}
			result->nRows++;
		}
		free(osqlda);

		if (fetch_stat != 100L){
			db->errcode=1; 
			return NULL;
		}
		if (isc_dsql_free_statement(db->status, &query_handle, DSQL_close)){
			db->errcode=1; 
			return NULL;
		}
		
	} else {
		/* Not select */
		if (isc_dsql_execute(db->status, &tr_handle, &query_handle, 1, osqlda)) {
			db->errcode=1;
			return NULL;
		}
		empty_res.nRows=0;
		empty_res.nCols=0;
		empty_res.items=NULL;
		empty_res.freeme=0;
		result=&empty_res;
	}
	if (isc_commit_transaction(db->status, &tr_handle)){
		db->errcode=1; 
		return NULL;
	}
	return result;
}
static iBASE_RES * sql_query(DB * db, char *query){
iBASE_RES * res;
	res=sql_ibase_query(db,query);
	if(db->errcode){
		displayError(db);
		if(strstr(db->errstr,"uplicat")){
			db->errcode=0;
		}
		return(NULL);
	}
	return(res);
}
/******************************** Template  ************************/
#else
typedef int UDB_RES;
typedef struct struct_db {
	int   *res;
	int   *res1;
	int connected;
	int errcode;
	char errstr[STRSIZ];
} DB;
#define SQL_NUM_ROWS(x)	0
#define SQL_FREE(x)
#define sql_value(x,y,z) 0
#define lock_dict(x)
#define	flush_dict(x)
#define unlock_dict(x)
#define lock_url(x)
#define unlock_url(x)
static int InitDB(DB *db){
	db->connected=1;return(0);
}
static int * sql_query(char *query){
	return(0);
}
#endif

/********************************************************************/
/* --------------------------------------------------------------- */
/* Almost Unified code for all supported backends                  */
/* --------------------------------------------------------------- */
/*******************************************************************/

static UDB_RES * urlres;
void * AllocDB(){
DB * db;
	db=(DB*)malloc(sizeof(DB));
#if (HAVE_IODBC || HAVE_UNIXODBC || HAVE_SOLID || HAVE_VIRT)
	db->hDbc=SQL_NULL_HDBC;
	db->hEnv=SQL_NULL_HENV;
	db->hstmt=SQL_NULL_HSTMT;
#endif
#if (HAVE_IBASE)
	db->DBH=NULL;
#endif
	db->res=NULL;
	db->res_limit=0;
	db->connected=0;
	db->errcode=0;
	db->errstr[0]=0;
	return((void*)db);
}
void FreeDB(void * db){
	if(db){
		CloseDB((DB*)db);
		free(db);
	}
}
int DBErrorCode(void *db){
	return(((DB*)db)->errcode);
}
char * DBErrorMsg(void *db){
	return(((DB*)db)->errstr);
}

void IncURLNumber(){
	LOCK_INDEXED();
	CurURLNumber++;
	UNLOCK_INDEXED();
}
int GetIndexedNumber(){
int result;
	LOCK_INDEXED();
	result=CurURLNumber;
	UNLOCK_INDEXED();
	return(result);
}
__INDLIB__ void SetMaxURLNumber(int n){
	LOCK_INDEXED();
	MaxURLNumber = n;
	UNLOCK_INDEXED();
}

/******************** Limits stuff *********************************/

__INDLIB__ int AddTagLimit(int tag){
#ifdef  HAVE_SQL_IN
	if(*tagstr)sprintf(STREND(tagstr)-1,",%d)",tag);
	else	sprintf(tagstr," AND url.tag IN (%d)",tag);
#else
	if(*tagstr)strcpy(STREND(tagstr)-1," OR ");
	else	strcat(tagstr," AND (");
	sprintf(STREND(tagstr),"url.tag=%d)",tag);
#endif
	return(0);
}
__INDLIB__ int AddStatusLimit(int status){
#ifdef  HAVE_SQL_IN
	if(*statusstr)sprintf(STREND(statusstr)-1,",%d)",status);
	else	sprintf(statusstr," AND url.status IN (%d)",status);
#else
	if(*statusstr)strcpy(STREND(statusstr)-1," OR ");
	else	strcat(statusstr," AND (");
	sprintf(STREND(statusstr),"url.status=%d)",status);
#endif
	return(0);
}
__INDLIB__ int AddURLLimit(char *URL){
	if(*urlstr)strcpy(STREND(urlstr)-1," OR ");
	else	strcat(urlstr," AND (");
	sprintf(STREND(urlstr),"url.url LIKE '%s')",URL);
	return(0);
}
__INDLIB__ int ClearLimits(){
	urlstr[0]=0;
	tagstr[0]=0;
	statusstr[0]=0;
	return(0);
}

/************* robots.txt stuff *********************************/
int  LoadRobots(INDEXER *Indexer){
int i,nrobots;
	LOCK_ROBOTS();
	if(Robots)FreeRobots(Indexer);
	((DB*)(Indexer->db))->res=sql_query(((DB*)(Indexer->db)),
		"SELECT hostinfo,path FROM robots");
	if(DBErrorCode(Indexer->db)){
		UNLOCK_ROBOTS();
		return(IND_ERROR);
	}
	nrobots=SQL_NUM_ROWS(((DB*)(Indexer->db))->res);
	if(nrobots){
		Robots=(ROBOT *)malloc((nrobots+1)*sizeof(ROBOT));
		for(i=0;i<nrobots;i++){
			Robots[i].hostinfo=strdup(sql_value(((DB*)(Indexer->db))->res,i,0));
			Robots[i].path=strdup(sql_value(((DB*)(Indexer->db))->res,i,1));
		}
		Robots[i].hostinfo=Robots[i].path=NULL;
	}
	SQL_FREE(((DB*)(Indexer->db))->res);
	UNLOCK_ROBOTS();
	return(IND_OK);
}
int DeleteRobotsFromHost(INDEXER *Indexer,char *hostinfo){
char qbuf[STRSIZ];
	sprintf(qbuf,"DELETE FROM robots WHERE hostinfo='%s'",hostinfo);
	sql_query(((DB*)(Indexer->db)),qbuf);
	if(DBErrorCode(Indexer->db))return(IND_ERROR);
	return(IND_OK);
}
int AddRobotsToHost(INDEXER *Indexer,char *hostinfo,char *s){
char qbuf[STRSIZ];
	sprintf(qbuf,"INSERT INTO robots (hostinfo,path) VALUES ('%s','%s')",hostinfo,s);
	sql_query(((DB*)(Indexer->db)),qbuf);
	if(DBErrorCode(Indexer->db))return(IND_ERROR);
	return(IND_OK);
}
int DeleteAllFromRobots(INDEXER *Indexer){
char qbuf[STRSIZ];
	sprintf(qbuf,"DELETE FROM robots");
	sql_query(((DB*)(Indexer->db)),qbuf);
	if(DBErrorCode(Indexer->db)){
		return(IND_ERROR);
	}
	return(IND_OK);
}


/************************* stopwords *******************************/

static int cmpstop(const void *s1,const void *s2){
	return(strcmp(((STOP_WORD*)s1)->word,((STOP_WORD*)s2)->word));
}
int  load_stop_list(void * db){
int i,j,cur=0;

	((DB*)db)->res=sql_query((DB*)db,"SELECT word,lang FROM stopword");
	if(((DB*)db)->errcode)return(IND_ERROR);

	nstoplist=SQL_NUM_ROWS(((DB*)db)->res);
	stoplist=(STOP_WORD *)malloc((nstoplist)*sizeof(STOP_WORD));
	for(i=0;i<nstoplist;i++){
		char *newword, *newlang;
		int found;

		newword=sql_value(((DB*)db)->res,i,0);
		newlang=sql_value(((DB*)db)->res,i,1);
		found=0;

		/* If the word is already in list     */
		/* We will not add it again           */
		/* But mark it as "international word"*/
		/* i.e. the word without language     */
		/* It will allow to avoid troubles    */
		/* with language guesser              */
		for(j=0;j<cur;j++){
			if(!strcmp(stoplist[j].word,newword)){
				stoplist[j].lang[0]=0;
				found=1;
				break;
			}
		}
		/* Add new stopword */
		if(!found){
			stoplist[cur].word=strdup(newword);
			strncpy(stoplist[cur].lang,newlang,2);
			stoplist[cur].lang[2]=0;
			cur++;
		}
	}
	SQL_FREE(((DB*)db)->res);
	/* If there was some inretnational words */
	/* We have to realloc memory to actually */
	/* required size                         */
	if(cur<nstoplist){
		nstoplist=cur;
		stoplist=(STOP_WORD *)realloc(stoplist,(nstoplist)*sizeof(STOP_WORD));
	}
	/* Sort it to run binary search later */
	qsort((void*)stoplist,nstoplist,sizeof(STOP_WORD),cmpstop);
	return(IND_OK);
}
STOP_WORD * IsStopWord(char *word){
int low = 0;
int high = nstoplist - 1;
int middle, match;

	if(!stoplist)return(0);
	while (low <= high) {
		middle = (low + high) / 2;
		match = strcmp(stoplist[middle].word, word);
		if (match == 0)
			return (&stoplist[middle]);
		if (match < 0)
			low = middle + 1;
		else
			high = middle - 1;
	}
	return(NULL);
}
/************************ URLs ***********************************/

int AddURL(INDEXER *Indexer,char *url,int referrer,int hops){
char *e_url;
int next_url_id=0;
char qbuf[STRSIZ]="AddURL";

	e_url=(char*)malloc(strlen(url)*2);escstr(e_url,url);

	switch(DBType){
	case DB_MSQL:
		((DB*)(Indexer->db))->res=sql_query((DB*)(Indexer->db),"SELECT _seq FROM url");
		if(DBErrorCode(Indexer->db)){
			free(e_url);
			return(IND_ERROR);
		}
		next_url_id=atoi(sql_value(((DB*)(Indexer->db))->res,0,0));
		SQL_FREE(((DB*)(Indexer->db))->res);
		sprintf(qbuf,"INSERT INTO url (url,referrer,hops,rec_id,crc,last_index_time,next_index_time,status) VALUES ('%s',%d,%d,%d,'',%d,%d,0)",e_url,referrer,hops,next_url_id,(int)now(),(int)now());
		break;
		
	case DB_SOLID:
	case DB_ORACLE:
		((DB*)(Indexer->db))->res=sql_query(Indexer->db,"SELECT next_url_id.nextval");
		if(DBErrorCode(Indexer->db)){
			free(e_url);
			return(IND_ERROR);
		}
		next_url_id=atoi(sql_value(((DB*)(Indexer->db))->res,0,0));
		SQL_FREE(((DB*)(Indexer->db))->res);
		sprintf(qbuf,"INSERT INTO url (url,referrer,hops,rec_id,crc,last_index_time,next_index_time,status) VALUES ('%s',%d,%d,%d,'',%d,%d,0)",e_url,referrer,hops,next_url_id,(int)now(),(int)now());
		break;
	default:
		sprintf(qbuf,"INSERT INTO url (url,referrer,hops,crc,last_index_time,next_index_time,status) VALUES ('%s',%d,%d,'',%d,%d,0)",e_url,referrer,hops,(int)now(),(int)now());
	}
	free(e_url);
	sql_query(((DB*)(Indexer->db)),qbuf);
	if(DBErrorCode(Indexer->db))
		return(IND_ERROR);
	else
		return(IND_OK);
}
int DeleteUrl(INDEXER *Indexer,int url_id){
char qbuf[STRSIZ];
int res;
	if((res=DeleteWordFromURL(Indexer,url_id))!=IND_OK)return(res);
	sprintf(qbuf,"DELETE FROM url WHERE rec_id=%d",url_id);
	sql_query(((DB*)(Indexer->db)),qbuf);
	if(DBErrorCode(Indexer->db))return(IND_ERROR);
	return(IND_OK);
}
int MarkForReindex(INDEXER *Indexer){
char qbuf[STRSIZ];
	LOCK_TARGET(Indexer);
	have_targets=1;
	UNLOCK_TARGET(Indexer);
	sprintf(qbuf,"UPDATE url SET next_index_time=%d WHERE rec_id>=0 %s%s%s",(int)now(),tagstr,urlstr,statusstr);
	sql_query(((DB*)(Indexer->db)),qbuf);
	if(DBErrorCode(Indexer->db))return(IND_ERROR);
	return(IND_OK);
}
int UpdateUrl(INDEXER *Indexer,int url_id,int status,int period){
char qbuf[STRSIZ];
	sprintf(qbuf,"UPDATE url SET status=%d,next_index_time=%d WHERE rec_id=%d",status,(int)(now()+period),url_id);
	sql_query(((DB*)(Indexer->db)),qbuf);
	if(DBErrorCode(Indexer->db))return(IND_ERROR);
	return(IND_OK);
}
int LongUpdateUrl(INDEXER *Indexer,
int url_id,int status,int changed,int size,int period,
int hint,int index,
char *last_modified_escaped,
char *text_escaped,
char *title_escaped,
char *content_type,
char *keywords_escaped,
char *descript_escaped,
char *digest,
char *lang){
char qbuf[STRSIZ];
char last_index_time[64]="";

	if(changed)sprintf(last_index_time,",last_index_time=%d",(int)now());
sprintf(qbuf,"\
UPDATE url SET \
status=%d,last_modified='%s',\
next_index_time=%d,\
tag=%d,txt='%s',title='%s',content_type='%s',docsize=%d,\
keywords='%s',description='%s',crc='%s',lang='%s' \
WHERE rec_id=%d",
	status,last_modified_escaped,(int)(now()+period),
	hint,text_escaped,title_escaped,content_type,size,
	keywords_escaped,descript_escaped,digest,lang,url_id);
	sql_query(((DB*)(Indexer->db)),qbuf);
	if(DBErrorCode(Indexer->db))return(IND_ERROR);
	return(IND_OK);
}


/********************** Words ***********************************/

int DeleteWordFromURL(INDEXER *Indexer,int url_id){
char qbuf[STRSIZ];
#ifdef USE_MULTIDICT
int i,last=0;
	for(i=MINDICT;i<MAXDICT;i++){
		if(last!=DICTNUM(i)){
			sprintf(qbuf,"DELETE FROM dict%d WHERE url_id=%d",
				DICTNUM(i),url_id);
			sql_query(((DB*)(Indexer->db)),qbuf);
			if(DBErrorCode(Indexer->db))return(IND_ERROR);
			last=DICTNUM(i);
		}
	}
#else
	sprintf(qbuf,"DELETE FROM dict WHERE url_id=%d",url_id);
	sql_query(((DB*)(Indexer->db)),qbuf);
	if(DBErrorCode(Indexer->db))return(IND_ERROR);
#endif
	return(IND_OK);
}

#ifdef USE_MULTIDICT
int StoreWords(INDEXER * Indexer,int url_id){
int i,wcur,len,res;
UDM_WORD *Word;
char qbuf[STRSIZ];
char tablename[64]="dict";

	wcur=Indexer->nwords;
	Word=Indexer->Word;
	if(IND_OK!=(res=DeleteWordFromURL(Indexer,url_id)))return(res);
	for(i=0;i<wcur;i++){
		if(Word[i].count){
			len=strlen(Word[i].word);
			len=DICTNUM(len);

			if(len)sprintf(tablename,"dict%d",len);
			/*
			if(Word[i].code)
			sprintf(qbuf,"INSERT INTO norm (url_id,word_id,intag) VALUES(%d,%d,%d)",url_id,Word[i].code,Word[i].count);
			else
			*/
			sprintf(qbuf,"INSERT INTO %s (url_id,word,intag) VALUES(%d,'%s',%d)",tablename,url_id,Word[i].word,Word[i].count);
			sql_query(((DB*)(Indexer->db)),qbuf);
			if(DBErrorCode(Indexer->db))return(IND_ERROR);
		}
	}
	return(IND_OK);
}
#else
static UDM_WORD * findword(int wcur,UDM_WORD *Word,char *w){
int i;
	for(i=0;i<wcur;i++)
		if(!strcmp(Word[i].word,w))
			return(&Word[i]);
	return(0);
}
int StoreWords(INDEXER * Indexer,int url_id){
int i,old,new,flush,wcur;
int were,added,deleted,updated;
char *s;UDM_WORD *w,*Word;
char *e;
char qbuf[STRSIZ];
	wcur=Indexer->nwords;
	Word=Indexer->Word;
	flush=were=added=deleted=updated=0;
	sprintf(qbuf,"SELECT word,intag FROM dict WHERE url_id=%d",url_id);
	((DB*)(Indexer->db))->res=sql_query(((DB*)(Indexer->db)),qbuf);
	if(DBErrorCode(Indexer->db))return(IND_ERROR);
	lock_dict((DB*)(Indexer->db));
	if(DBErrorCode(Indexer->db))return(IND_ERROR);

	were=SQL_NUM_ROWS(((DB*)(Indexer->db))->res);
	
	for(i=0;i<were;i++){
		e=s=sql_value(((DB*)(Indexer->db))->res,i,0);
		old=atoi(sql_value(((DB*)(Indexer->db))->res,i,1));
		rtrim_chars(s," ");
		if((w=findword(wcur,Word,s))){
			new=w->count;
			if((new)&&(old)&&(new!=old)){
				sprintf(qbuf,"UPDATE dict SET intag=%d WHERE word='%s' AND url_id='%d'",new,s,url_id);
				sql_query(((DB*)(Indexer->db)),qbuf);
				if(DBErrorCode(Indexer->db))return(IND_ERROR);
				updated++;
				flush++;
			}
			w->count=0;
		}else{
			sprintf(qbuf,"DELETE FROM dict WHERE url_id=%d AND word='%s'",url_id,s);
			sql_query(((DB*)(Indexer->db)),qbuf);
			if(DBErrorCode(Indexer->db))return(IND_ERROR);
			deleted++;flush++;
		}
		if(flush>1024){
			flush_dict(Indexer->db);
			flush=0;
		}
	}
	SQL_FREE(((DB*)(Indexer->db))->res);
	for(i=0;i<wcur;i++){
		if(Word[i].count){
			sprintf(qbuf,"INSERT INTO dict (url_id,word,intag) VALUES(%d,'%s',%d)",url_id,Word[i].word,Word[i].count);
			sql_query(((DB*)(Indexer->db)),qbuf);
			if(DBErrorCode(Indexer->db))return(IND_ERROR);
			flush++;added++;
			if(flush>1024){
				flush_dict((DB*)(Indexer->db));
				flush=0;
			}
		}
	}
	unlock_dict((DB*)(Indexer->db));
	if(DBErrorCode(Indexer->db))return(IND_ERROR);
	return(IND_OK);
}
#endif

int DeleteAllFromUrl(INDEXER *Indexer){
char qbuf[STRSIZ];
	sprintf(qbuf,"DELETE FROM url");
	sql_query(((DB*)(Indexer->db)),qbuf);
	if(DBErrorCode(Indexer->db))return(IND_ERROR);
	return(IND_OK);
}
int DeleteAllFromDict(INDEXER *Indexer){
char qbuf[STRSIZ];

#ifdef USE_MULTIDICT
int i,last=0;
	for(i=MINDICT;i<MAXDICT;i++){
		if(last!=DICTNUM(i)){
			sprintf(qbuf,"DELETE FROM dict%d",DICTNUM(i));
			sql_query(((DB*)(Indexer->db)),qbuf);
			if(DBErrorCode(Indexer->db))return(IND_ERROR);
			last=DICTNUM(i);
		}
	}
#else
	sprintf(qbuf,"DELETE FROM dict");
	sql_query(((DB*)(Indexer->db)),qbuf);
	if(DBErrorCode(Indexer->db))return(IND_ERROR);
#endif
	return(IND_OK);
}


/************************ Clones stuff ***************************/
int FindOrigin(INDEXER *Indexer,char *crc){
int i=0,origin=0;
char *o;
char qbuf[STRSIZ];
	if(!crc)return(0);if(!(*crc))return(0);
	sprintf(qbuf,"SELECT rec_id FROM url WHERE crc='%s'",crc);
	((DB*)(Indexer->db))->res=sql_query(((DB*)(Indexer->db)),qbuf);
	if(DBErrorCode(Indexer->db))return(0);
	for(i=0;i<SQL_NUM_ROWS(((DB*)(Indexer->db))->res);i++)
		if((o=sql_value(((DB*)(Indexer->db))->res,i,0)))
			if((!origin)||(origin>atoi(o)))
				origin=atoi(o);
	SQL_FREE(((DB*)(Indexer->db))->res);
	return(origin);
}
int  UpdateClone(INDEXER *Indexer,int url_id,int status,int period,
char *content_type,char *last_modified_escaped,char *digest){
char qbuf[STRSIZ];
	sprintf(qbuf,
"UPDATE url \
SET crc='%s',status=%d,content_type='%s',\
last_modified='%s',next_index_time=%d \
WHERE rec_id=%d",digest,status,content_type,\
last_modified_escaped,(int)(now()+period),url_id);
	sql_query(((DB*)(Indexer->db)),qbuf);
	if(DBErrorCode(Indexer->db))return(IND_ERROR);
	return(IND_OK);
}

/*******************************************************************/

DOCUMENT *GetDocInfo(INDEXER *Indexer,int index_flags){
char urlin[STRSIZ]="";
char sortstr[64]="";
int i=0,url_num;
DOCUMENT * Result=NULL;
char qbuf[STRSIZ];

	LOCK_STARTUP(Indexer);
	if(!index_started){
		if(!mark_done){
			MarkForReindex(Indexer);
			if(DBErrorCode(Indexer->db)){
				UNLOCK_STARTUP(Indexer);
				return(NULL);
			}
		}
		LoadRobots(Indexer);
		if(DBErrorCode(Indexer->db)){
			UNLOCK_STARTUP(Indexer);
			return(NULL);
		}
		load_stop_list(Indexer->db);
		if(DBErrorCode(Indexer->db)){
			UNLOCK_STARTUP(Indexer);
			return(NULL);
		}
		index_started=1;
	}
	UNLOCK_STARTUP(Indexer);

#ifdef HAVE_PTHREAD
	LOCK_HREFS(Indexer);
	LOCK_TARGET(Indexer);
	i=have_targets||nhrefs;
	UNLOCK_TARGET(Indexer);
	UNLOCK_HREFS(Indexer);
	if(!i){
		return(NULL);
	}
#endif

	if((GetIndexedNumber()>=MaxURLNumber)&&(MaxURLNumber>=0)){
		StoreHrefs(Indexer);
		return(NULL);
	}

	LOCK_DOCS(Indexer);
	if(currow>=nrows){
		StoreHrefs(Indexer);
		if(DBErrorCode(Indexer->db)){
			UNLOCK_DOCS(Indexer);
			return(NULL);
		}
		if(urlres)SQL_FREE(urlres);
		lock_url((DB*)(Indexer->db));

		url_num=URL_SELECT_CACHE;
		if(MaxURLNumber>0){
			url_num=MaxURLNumber-GetIndexedNumber();
			if(url_num>URL_SELECT_CACHE)
				url_num=URL_SELECT_CACHE;
		}

		sprintf(sortstr,(index_flags&FLAG_EXP_FIRST)?" ORDER BY next_index_time":"");
#ifdef HAVE_SQL_LIMIT
		sprintf(qbuf,
		"SELECT url,rec_id,docsize,status,last_index_time,hops,crc,last_modified FROM url WHERE next_index_time<=%d %s%s%s%s LIMIT %d",
			(int)now(),tagstr,urlstr,statusstr,sortstr,url_num);
#else
		((DB*)(Indexer->db))->res_limit=url_num;
		sprintf(qbuf,
		"SELECT url,rec_id,docsize,status,last_index_time,hops,crc,last_modified FROM url WHERE next_index_time<=%d %s%s%s%s",
			(int)now(),tagstr,urlstr,statusstr,sortstr);
#endif
		urlres=sql_query(((DB*)(Indexer->db)),qbuf);
		if(DBErrorCode(Indexer->db)){
			UNLOCK_DOCS(Indexer);
			return(NULL);
		}
		nrows=SQL_NUM_ROWS(urlres);
		currow=0;
		if(!nrows){
			LOCK_TARGET(Indexer);
			have_targets=0;
			UNLOCK_TARGET(Indexer);
			unlock_url((DB*)(Indexer->db));
			UNLOCK_DOCS(Indexer);
			return(NULL);
		}
		urlin[0]=0;

		for(i=0;i<nrows;i++){
#ifdef  HAVE_SQL_IN
			if(urlin[0])strcat(urlin,",");
			strcat(urlin,sql_value(urlres,i,1));
#else
			sprintf(qbuf,"UPDATE url SET next_index_time=%d WHERE rec_id=%d",(int)(now()+3600),
				atoi(sql_value(urlres,i,1)));
			sql_query(((DB*)(Indexer->db)),qbuf);
			if(DBErrorCode(Indexer->db)){
				UNLOCK_DOCS(Indexer);
				return(NULL);
			}
#endif
		}
#ifdef HAVE_SQL_IN
		sprintf(qbuf,"UPDATE url SET next_index_time=%d WHERE rec_id in (%s)",(int)(now()+3600),urlin);	
		sql_query(((DB*)(Indexer->db)),qbuf);
		if(DBErrorCode(Indexer->db)){
			UNLOCK_DOCS(Indexer);
			return(NULL);
		}
#endif
		unlock_url((DB*)(Indexer->db));
	}
	Result=(DOCUMENT *)malloc(sizeof(DOCUMENT));
	Result->content_type=NULL;
	Result->title=NULL;
	Result->text=NULL;
	Result->last_index_time=NULL;
	Result->next_index_time=NULL;
	Result->keywords=NULL;
	Result->description=NULL;
	Result->crc=NULL;

	Result->url=strdup(sql_value(urlres,currow,0));
	Result->url_id=atoi(sql_value(urlres,currow,1));
	Result->size=atoi(sql_value(urlres,currow,2));
	Result->status=atoi(sql_value(urlres,currow,3));
	Result->last_index_time=strdup(sql_value(urlres,currow,4));
	Result->hops=atoi(sql_value(urlres,currow,5));
	Result->crc=strdup(sql_value(urlres,currow,6))?
		strdup(sql_value(urlres,currow,6)):strdup("___");
	Result->last_modified=strdup(sql_value(urlres,currow,7));

	currow++;
	UNLOCK_DOCS(Indexer);
	return(Result);
}
void FreeDoc(DOCUMENT *Result){
	if(!Result)return;
	free_it(Result->url);
	free_it(Result->content_type);
	free_it(Result->last_modified);
	free_it(Result->title);
	free_it(Result->text);
	free_it(Result->last_index_time);
	free_it(Result->next_index_time);
	free_it(Result->keywords);
	free_it(Result->description);
	free_it(Result->crc);
	free(Result);
}


/************************************************************/
/* Misc functions                                           */
/************************************************************/

STAT * GetStatistics(INDEXER *Indexer){
int i,j;
STAT *stat=NULL;
char qbuf[STRSIZ];

#if (HAVE_MYSQL|HAVE_PGSQL)
#if   (HAVE_MYSQL)
	sprintf(qbuf,"SELECT status,sum(next_index_time<=%d),count(*) FROM url WHERE 1=1 %s%s%s GROUP BY status",(int)now(),tagstr,urlstr,statusstr);
#elif (HAVE_PGSQL)
	sprintf(qbuf,"SELECT status,sum(case when next_index_time<=%d then 1 else 0 end),count(*) FROM url WHERE 1=1 %s%s%s GROUP BY status",(int)now(),tagstr,urlstr,statusstr);
#endif
	((DB*)(Indexer->db))->res=sql_query(((DB*)(Indexer->db)),qbuf);
	if(DBErrorCode(Indexer->db))return(NULL);
	if((j=SQL_NUM_ROWS(((DB*)(Indexer->db))->res))){
		stat=(STAT *)malloc(sizeof(STAT)*(j+1));
		for(i=0;i<j;i++){
			stat[i].status=atoi(sql_value(((DB*)(Indexer->db))->res,i,0));
			stat[i].expired=atoi(sql_value(((DB*)(Indexer->db))->res,i,1));
			stat[i].total=atoi(sql_value(((DB*)(Indexer->db))->res,i,2));
		}
		stat[i].status=-1;
	}
	SQL_FREE(((DB*)(Indexer->db))->res);
#elif (HAVE_IBASE||HAVE_MSQL || HAVE_IODBC || HAVE_UNIXODBC || HAVE_SOLID || HAVE_VIRT)
	sprintf(qbuf,"SELECT status,next_index_time FROM url WHERE rec_id>=0 %s%s%s",tagstr,urlstr,statusstr);
	((DB*)(Indexer->db))->res=sql_query(((DB*)(Indexer->db)),qbuf);
	if(DBErrorCode(Indexer->db))return(NULL);
	if(SQL_NUM_ROWS(((DB*)(Indexer->db))->res)){
		stat=(STAT *)malloc(sizeof(STAT)*(SQL_NUM_ROWS(((DB*)(Indexer->db))->res)+1));
		stat[0].status=-1;
		for(i=0;i<SQL_NUM_ROWS(((DB*)(Indexer->db))->res);i++){
			int n;
			n=now();
			for(j=0;stat[j].status!=-1;j++){
				if(stat[j].status==atoi(sql_value(((DB*)(Indexer->db))->res,i,0))){
					if(atoi(sql_value(((DB*)(Indexer->db))->res,i,1))<=n)
						stat[j].expired++;
					stat[j].total++;
					break;
				}
			}
			if(stat[j].status==-1){
				stat[j].status=atoi(sql_value(((DB*)(Indexer->db))->res,i,0));
				stat[j].expired=0;
				if(atoi(sql_value(((DB*)(Indexer->db))->res,i,1))<=n)
					stat[j].expired++;
				stat[j].total=1;
				stat[j+1].status=-1;
			}
		}
	}
	SQL_FREE(((DB*)(Indexer->db))->res);
#endif
	return(stat);
}


__INDLIB__ int ShowReferers(INDEXER *Indexer){
int i,j;
char str[STRSIZ];
char qbuf[STRSIZ];
#if (WIN32|WINNT)
#else
	printf("\n          URLs and referers \n\n");
#endif
	sprintf(qbuf,"SELECT url.status,url2.url,url.url FROM url,url url2 WHERE url.referrer=url2.rec_id %s%s%s ",tagstr,urlstr,statusstr);
	((DB*)(Indexer->db))->res=sql_query(((DB*)(Indexer->db)),qbuf);
	if(DBErrorCode(Indexer->db))return(IND_ERROR);
	j=SQL_NUM_ROWS(((DB*)(Indexer->db))->res);
	for(i=0;i<j;i++){
		sprintf(str,"%s %s <- %s",
			sql_value(((DB*)(Indexer->db))->res,i,0),
			sql_value(((DB*)(Indexer->db))->res,i,2),
			sql_value(((DB*)(Indexer->db))->res,i,1));
		REFERERINFO(Indexer->handle,str);
	}
	SQL_FREE(((DB*)(Indexer->db))->res);
	return(IND_OK);
}

int ClearDB(INDEXER *Indexer){
int i,j,err;
char qbuf[STRSIZ];
	if(!tagstr[0]&&!urlstr[0]&&!statusstr[0]){
#if (WIN32|WINNT)
#else
		printf("Deleting all records\n");
#endif
		if((IND_OK!=(err=DeleteAllFromDict(Indexer))))return(err);
		if((IND_OK!=(err=DeleteAllFromUrl(Indexer))))return(err);
		if((IND_OK!=(err=DeleteAllFromRobots(Indexer))))return(err);
	}else{
#if (WIN32|WINNT)
#else
		printf("Deleting part\n");
#endif
		j=0;
		while(1){
			sprintf(qbuf,"SELECT rec_id FROM url WHERE rec_id>=0 %s%s%s LIMIT %d",tagstr,urlstr,statusstr,URL_DELETE_CACHE);
			((DB*)(Indexer->db))->res=sql_query(((DB*)(Indexer->db)),qbuf);
			if(DBErrorCode(Indexer->db))return(IND_ERROR);
			if(SQL_NUM_ROWS(((DB*)(Indexer->db))->res)){
#ifdef HAVE_SQL_IN
				char urlin[STRSIZ]="";
				urlin[0]=0;
				for(i=0;i<SQL_NUM_ROWS(((DB*)(Indexer->db))->res);i++){
					if(i)strcat(urlin,",");
					strcat(urlin,sql_value(((DB*)(Indexer->db))->res,i,0));
				}
				j+=i;
#ifdef USE_MULTIDICT
				{
				int last=0;
				for(i=MINDICT;i<MAXDICT;i++){
					if(last!=DICTNUM(i)){
						sprintf(qbuf,"DELETE FROM dict%d WHERE url_id in (%s)",
							DICTNUM(i),urlin);
						sql_query(((DB*)(Indexer->db)),qbuf);
						if(DBErrorCode(Indexer->db))return(IND_ERROR);
						last=DICTNUM(i);
					}
				}
				}
#else
				sprintf(qbuf,"DELETE FROM dict WHERE url_id in (%s)",urlin);
				sql_query(((DB*)(Indexer->db)),qbuf);
				if(DBErrorCode(Indexer->db))return(IND_ERROR);
#endif
				sprintf(qbuf,"DELETE FROM url WHERE rec_id in (%s)",urlin);
				sql_query(((DB*)(Indexer->db)),qbuf);
				if(DBErrorCode(Indexer->db))return(IND_ERROR);

				SQL_FREE(((DB*)(Indexer->db))->res);
#else
				for(i=0;i<SQL_NUM_ROWS(((DB*)(Indexer->db))->res);i++)
					if(IND_OK!=DeleteUrl(Indexer,atoi(sql_value(((DB*)(Indexer->db))->res,i,0))))
						return(IND_ERROR);
				j+=i;
#endif
				sprintf(qbuf,"%d",j);
				DELETEINFO(Indexer->handle,qbuf);
				UDMFLASH(stdout);
			}else{
				SQL_FREE(((DB*)(Indexer->db))->res);
				break;
			}
		}
	}
#if (WIN32|WINNT)
#else
	printf("Done\n");
#endif
	return(IND_OK);
}

/******************* Search stuff ************************************/

DOCUMENT * CloneList(void * db,char *crc){
int i;
DOCUMENT * r=NULL;
char qbuf[STRSIZ];
	sprintf(qbuf,"SELECT rec_id,url,content_type,last_modified FROM url WHERE crc='%s'",crc);
	((DB*)db)->res=sql_query((DB*)db,qbuf);
	if(DBErrorCode(db))return(NULL);
	r=(DOCUMENT*)malloc(sizeof(DOCUMENT)*(SQL_NUM_ROWS(((DB*)db)->res)+1));
	for(i=0;i<SQL_NUM_ROWS(((DB*)db)->res);i++){
		r[i].url_id=atoi(sql_value(((DB*)db)->res,i,0));
		r[i].url=strdup(sql_value(((DB*)db)->res,i,1));
		r[i].content_type=strdup(sql_value(((DB*)db)->res,i,2));
		r[i].last_modified=strdup(sql_value(((DB*)db)->res,i,3));
		r[i+1].url_id=0;
	}
	SQL_FREE(((DB*)db)->res);
	return(r);
}

static int AddInList(char *w,char **list){
char **i;
	i=list;
	while(*i){
		if(!strcmp(*i,w))return(0);
		i++;
	}
	*i=strdup(w);i++;*i=0;return(0);
}

char ** GetAllWordForms(void * db,char *word){
int l;
char **list=NULL;

	list=(char **)malloc((sizeof(char **))*128);*list=NULL;
	l=strlen(word);AddInList(word,list);
	return(list);
}
/********************************************************/

typedef struct wrd{
	int  url_id;
	int  count;
	int  weight;
} SEARCHWORD;

static int cmpword(const void *s1,const void *s2){
int res;
	if(!(res=(((SEARCHWORD*)s2)->count)-(((SEARCHWORD*)s1)->count)))
		if(!(res=(((SEARCHWORD*)s2)->weight-(((SEARCHWORD*)s1)->weight))))
			if(!(res=(((SEARCHWORD*)s1)->url_id-(((SEARCHWORD*)s2)->url_id))));
	return(res);
}
static int cmpurlid(const void *s1,const void *s2){
	return(((SEARCHWORD*)s2)->url_id-((SEARCHWORD*)s1)->url_id);
}

DOCUMENT * Find(void * db,char *q,int np,int ps,int search_mode,int sort_order,char *wordinfo,int *found){
DOCUMENT * r=NULL;
char *w;
int i,j,num_words=0,first,last,num;
SEARCHWORD * wrd=NULL;
char qbuf[STRSIZ];
char *lasttok;
int url_id;
int weight=0;
int nwrd=0;
int curnum=0;
#ifdef HAVE_MYSQL
MYSQL_ROW row;
#endif

	load_stop_list(db);
	*wordinfo=0;
	if(DBErrorCode(db))return(NULL);
	w=gettoken(q,SEPARATOR,&lasttok);
	while(w){
#ifdef USE_MULTIDICT
		int len;
#endif
		char ** ww;
		char *rw;
		char tablename[32]="dict";

		ww=NormalizeWord(w);
		rw=ww?*ww:w;

		if(IsStopWord(rw)){
			sprintf(STREND(wordinfo)," %s :stopword",rw);
			w=gettoken(NULL,SEPARATOR,&lasttok);
			continue;
		}
		num_words++;
#ifdef USE_MULTIDICT
		len=strlen(rw);len=DICTNUM(len);
		sprintf(tablename,"dict%d",len);
#endif
		if(*tagstr||*statusstr||*urlstr)
		
			sprintf(qbuf,"\
SELECT %s.url_id,%s.intag \
FROM %s,url \
WHERE %s.word='%s' \
AND url.rec_id=%s.url_id %s%s%s",
			tablename,tablename,tablename,tablename,
			rw,tablename,tagstr,statusstr,urlstr);
		else
			sprintf(qbuf,"\
SELECT url_id,intag \
FROM %s \
WHERE word='%s'",
			tablename,rw);

		((DB*)db)->res=sql_query((DB*)db,qbuf);
		if(DBErrorCode(db))return(NULL);
		num=SQL_NUM_ROWS(((DB*)db)->res);
		sprintf(STREND(wordinfo)," %s : %d",rw,num);
		
		if(!nwrd){
			nwrd=num;
			wrd=(SEARCHWORD*)malloc(nwrd*sizeof(SEARCHWORD));
		}else{
			nwrd+=num;
			wrd=(SEARCHWORD*)realloc(wrd,nwrd*sizeof(SEARCHWORD));
		}
		
		for(i=0;i<num;i++){
#ifdef HAVE_MYSQL
			/* mysql_data_seek is slow */
			/* We will use sequential fetch instead*/
			row=mysql_fetch_row(((DB*)db)->res);
			url_id=atoi(row[0]);
			weight=atoi(row[1]);
#else
			url_id=atoi(sql_value(((DB*)db)->res,i,0));
			weight=atoi(sql_value(((DB*)db)->res,i,1));
#endif
			wrd[curnum].url_id=url_id;
			wrd[curnum].count=1;
			wrd[curnum].weight=weight;
			curnum++;
		}
		SQL_FREE(((DB*)db)->res);
		w=gettoken(NULL,SEPARATOR,&lasttok);
	}
	if(!nwrd)return(0);
	qsort((void*)wrd,nwrd,sizeof(SEARCHWORD),cmpurlid);
	j=0;
	for(i=1;i<nwrd;i++){
		if(wrd[j].url_id==wrd[i].url_id){
			wrd[j].weight+=wrd[i].weight;
			wrd[j].count++;
		}else{
			j++;
			wrd[j].url_id=wrd[i].url_id;
			wrd[j].count=wrd[i].count;
			wrd[j].weight=wrd[i].weight;
		}
	}
	nwrd=j+1;
	
	qsort((void*)wrd,nwrd,sizeof(SEARCHWORD),cmpword);
	if(search_mode==MOD_ALL){
		for(i=0;i<nwrd;i++)
			if(wrd[i].count<num_words){
				nwrd=i;
				break;
			}
	}
	if(!nwrd)return(0);
	*found=nwrd;
	first=np*ps;last=first+ps;
	if(last>nwrd)last=nwrd;
	nwrd=last-first;
	r=(DOCUMENT*)malloc(sizeof(DOCUMENT)*(nwrd+1));
	for(i=0;i<nwrd;i++){

		r[nwrd].url_id=0;
		sprintf(qbuf,"SELECT \
rec_id,url,content_type,last_modified,title,txt, \
docsize,last_index_time,next_index_time, \
referrer,keywords,description,crc FROM url \
WHERE rec_id=%d",wrd[i+first].url_id);
		((DB*)db)->res=sql_query(((DB*)db),qbuf);
		if(DBErrorCode(db))return(NULL);
		if(SQL_NUM_ROWS(((DB*)db)->res)){
			r[i].url_id=wrd[i+first].url_id;
			r[i].rating=wrd[i+first].count;
			r[i].url=strdup(sql_value(((DB*)db)->res,0,1));
			r[i].content_type=strdup(sql_value(((DB*)db)->res,0,2));
			r[i].last_modified=strdup(sql_value(((DB*)db)->res,0,3));
			r[i].title=strdup(sql_value(((DB*)db)->res,0,4));
			r[i].text=strdup(sql_value(((DB*)db)->res,0,5));
			r[i].size=atoi(sql_value(((DB*)db)->res,0,6));
			r[i].last_index_time=strdup(sql_value(((DB*)db)->res,0,7));
			r[i].next_index_time=strdup(sql_value(((DB*)db)->res,0,8));
			r[i].referrer=atoi(sql_value(((DB*)db)->res,0,9));
			r[i].keywords=strdup(sql_value(((DB*)db)->res,0,10));
			r[i].description=strdup(sql_value(((DB*)db)->res,0,11));
			r[i].crc=strdup(sql_value(((DB*)db)->res,0,12));
		}
		SQL_FREE(((DB*)db)->res);
	}
	r[nwrd].url_id=0;
	return(r);
}


/***********************************************************/

#ifdef USE_HTDB
static char * get_path_part(char *path,char *dst,int part){
char *s,*e;
int i=0;

	s=path;
	while(*s){
		if(i==part){
			if((e=strchr(s,'/')))strncpy(dst,s,e-s);
			else	strcpy(dst,s);
			return(dst);
		}
		if(*s=='/')i++;
		s++;
	}
	*dst=0;
	return(dst);
}

static char * include_params(char *src,char *path,char *dst){
char *s,*e;
int i;
	s=src;e=dst;*e=0;
	while(*s){
		if((*s=='\\')){
			*e=*(s+1);e++;*e=0;s+=2;
			continue;
		}
		if(*s!='$'){
			*e=*s;
			s++;e++;*e=0;
			continue;
		}
		s++;i=atoi(s);
		while((*s>='0')&&(*s<='9'))s++;
		get_path_part(path,e,i);
		while(*e)e++;
	}
	return(dst);
}


int HTDBGet(INDEXER * Indexer,char *path,char *filename,char * htdb_list, char * htdb_doc,
		char *buf,int MaxDocSize){
int i;
char qbuf[STRSIZ]="";
DB * db;


	db=(DB*)Indexer->db;
	buf[0]=0;
	if(*filename){
		char real_path[STRSIZ]="";
		sprintf(real_path,"%s%s",path,filename);
		if(!htdb_doc)htdb_doc="";
		include_params(htdb_doc,real_path,qbuf);
		db->res=sql_query(db,qbuf);
		if(DBErrorCode(db))return(0);
		if(SQL_NUM_ROWS(db->res)==1){
			sprintf(buf,"%s",sql_value(db->res,0,0));
		}else{
			sprintf(buf,"HTTP/1.0 404 Not Found\r\n\r\n");
		}
		SQL_FREE(db->res);
	}else{
		include_params(htdb_list,path,qbuf);
		db->res=sql_query(db,qbuf);
		if(DBErrorCode(db))return(0);
		sprintf(buf,"HTTP/1.0 200 OK\r\nContent-type: text/html\r\n\r\n<HTML><BODY>\n");
		for(i=0;i<SQL_NUM_ROWS(db->res);i++){
			sprintf(STREND(buf),"<a href=\"%s\">%s</a><br>\n",
				sql_value(db->res,i,0),
				sql_value(db->res,i,0));
		}
		SQL_FREE(db->res);
		strcat(buf,"</BODY></HTML>\n");
	}
	return(strlen(buf));
}
#endif
